﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.UI.Xaml.Controls;

namespace SimplePuzzle
{
    /// <summary>
    /// This class controls the puzzle game
    /// </summary>
    public class GameController
    {
        // --- Variables storing the state of the game
        private Canvas _tableCanvas;
        private PuzzlePiece[,] _puzzles;
        private int _startEmptyPosition;
        private PuzzlePiece[,] _startPuzzles;

        /// <summary>
        /// Gets the width of the puzzle table
        /// </summary>
        public int Width { get; set; }

        /// <summary>
        /// Gets the height of the puzzle table
        /// </summary>
        public int Height { get; set; }

        /// <summary>
        /// Gets the currently empty position of the table
        /// </summary>
        public int EmptyPosition { get; private set; }

        /// <summary>
        /// Gets the state of the current game
        /// </summary>
        public GameState State { get; private set; }

        /// <summary>
        /// Raised when a new game has been started
        /// </summary>
        public event EventHandler GameStarted;

        /// <summary>
        /// Raised when the shuffle of the puzzle has been started
        /// </summary>
        public event EventHandler ShuffleStarted;

        /// <summary>
        /// Raised when the shuffle of the puzzle is completed
        /// </summary>
        public event EventHandler ShuffleCompleted;

        /// <summary>
        /// Raised when the game is completed, because the puzzles are ordered.
        /// </summary>
        public event EventHandler GameCompleted;
        
        /// <summary>
        /// Creates a new game controller using the specified canvas
        /// </summary>
        /// <param name="canvas">Canvas to show the puzzle pieces</param>
        public GameController(Canvas canvas)
        {
            _tableCanvas = canvas;
            _tableCanvas.Children.Clear();
        }

        /// <summary>
        /// Starts a new game with the specified height and width
        /// </summary>
        /// <param name="height">Table height</param>
        /// <param name="width">Table width</param>
        public async void StartNew(int height, int width)
        {
            Width = width;
            Height = height;
            InitGameTable();
            if (GameStarted != null) GameStarted(this, EventArgs.Empty);
            State = GameState.Shuffle;
            await Shuffle(10);
            //await Shuffle(height * width * 10);
            SaveStartPositions();
            State = GameState.InProgress;
        }

        /// <summary>
        /// Initializes the table with puzzles
        /// </summary>
        private void InitGameTable()
        {
            _puzzles = new PuzzlePiece[Height, Width];
            _tableCanvas.Children.Clear();
            _tableCanvas.Width = Width * PuzzlePiece.PUZZLE_WIDTH;
            _tableCanvas.Height = Height * PuzzlePiece.PUZZLE_HEIGTH;

            for (var i = 0; i < Height * Width - 1; i++)
            {
                var newPiece = new PuzzlePiece(this, i);
                newPiece.SetPosition(i / Width, i % Width);
                this[i] = newPiece;
                _tableCanvas.Children.Add(newPiece);
            }
            EmptyPosition = Height * Width - 1;
            this[EmptyPosition] = null;
        }

        /// <summary>
        /// Saves the start positions for later reset
        /// </summary>
        private void SaveStartPositions()
        {
            _startPuzzles = new PuzzlePiece[Height, Width];
            for (int i = 0; i < Height; i++)
                for (int j = 0; j < Width; j++ )
                    _startPuzzles[i, j] = _puzzles[i, j];
            _startEmptyPosition = EmptyPosition;
        }

        /// <summary>
        /// Restores the saved start positions
        /// </summary>
        public void RestoreStartPositions()
        {
            for (int i = 0; i < Height; i++)
                for (int j = 0; j < Width; j++)
                {
                    var puzzle = _puzzles[i, j] = _startPuzzles[i, j];
                    if (puzzle != null) puzzle.SetPosition(i, j);
                }
            EmptyPosition = _startEmptyPosition;
        }

        /// <summary>
        /// Gets the puzzle piece in the specified row and column
        /// </summary>
        /// <param name="row">Row position</param>
        /// <param name="column">Column pozition</param>
        /// <returns></returns>
        public PuzzlePiece this[int row, int column]
        {
            get { return _puzzles[row, column]; }
            set { _puzzles[row, column] = value; }
        }

        /// <summary>
        /// Gets the puzzle in the specified position
        /// </summary>
        /// <param name="position">Puzzle position</param>
        /// <returns></returns>
        public PuzzlePiece this[int position]
        {
            get { return _puzzles[position / Width, position % Width]; }
            set { _puzzles[position / Width, position % Width] = value; }
        }

        /// <summary>
        /// Gets the poisitons a puzzle can be moved to from the specified source position
        /// </summary>
        /// <param name="position">Source position</param>
        /// <returns>List of available target positions</returns>
        public IEnumerable<int> TargetPositionsFrom(int position)
        {
            var row = position / Width;
            var col = position % Width;
            if (row > 0) yield return (row - 1) * Width + col;
            if (row < Height - 1) yield return (row + 1) * Width + col;
            if (col > 0) yield return row * Width + col - 1;
            if (col < Width - 1) yield return row * Width + col + 1;
        }

        /// <summary>
        /// Checks whether the puzzle in the specified position can be moved.
        /// </summary>
        /// <param name="position"></param>
        /// <returns></returns>
        public bool CanMove(int position)
        {
            return TargetPositionsFrom(EmptyPosition).Any(p => p == position);
        }

        /// <summary>
        /// Moves the puzzle in the specified position and optionally checks 
        /// the completion of the puzzle game
        /// </summary>
        /// <param name="position">Source puzzle position</param>
        /// <param name="checkComplete">Signs if completeness should be checked</param>
        public void MakeMove(int position, bool checkComplete = true)
        {
            if (CanMove(position))
            {
                var row = EmptyPosition / Width;
                var col = EmptyPosition % Width;
                var movingPuzzle = this[position];
                movingPuzzle.SetPosition(row, col);
                this[EmptyPosition] = movingPuzzle;
                this[position] = null;
                EmptyPosition = position;
                if (checkComplete && IsCompleted)
                {
                    State = GameState.Completed;
                    if (GameCompleted != null)
                    {
                        GameCompleted(this, EventArgs.Empty);
                    }
                }
            }
        }

        /// <summary>
        /// Gets the flag indicating if game is completed
        /// </summary>
        public bool IsCompleted
        {
            get 
            {
                for (var i = 0; i < Height * Width - 1; i++)
                {
                    if (this[i] == null || this[i].PuzzleId != i) return false;
                }
                return true;
            }
        }

        /// <summary>
        /// Shuffles the puzzles on the table
        /// </summary>
        /// <param name="count"></param>
        /// <returns></returns>
        public async Task Shuffle(int count)
        {
            if (ShuffleStarted != null) ShuffleStarted(this, EventArgs.Empty);
            var rnd = new Random((int)DateTime.Now.Ticks);
            for (int i = 0; i < count; i++)
            {
                var newPositions = TargetPositionsFrom(EmptyPosition).ToList();
                var index = rnd.Next(newPositions.Count);
                MakeMove(newPositions[index], false);
                await Task.Delay(50);
            }
            if (ShuffleCompleted != null) ShuffleCompleted(this, EventArgs.Empty);
        }
    }

    /// <summary>
    /// Shows the possible states of the game.
    /// </summary>
    public enum GameState
    {
        New,
        Shuffle,
        InProgress,
        Completed
    }
}
