﻿using System.Linq;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Input;

namespace SimplePuzzle
{
    /// <summary>
    /// This user control displays a simple puzzle piece.
    /// </summary>
    public sealed partial class PuzzlePiece : UserControl
    {
        /// <summary>
        /// Width and height of a puzzle in screen coordinates
        /// </summary>
        public const int PUZZLE_WIDTH = 125;
        public const int PUZZLE_HEIGTH = 125;

        // --- The Id of the puzzle that identifies its final position
        private int _puzzleId;

        /// <summary>
        /// Gets the table manmaging this puzzle
        /// </summary>
        public GameController Table { get; set; }

        /// <summary>
        /// Gets the row position of this puzzle
        /// </summary>
        public int Row { get; private set; }

        /// <summary>
        /// Gets the column position of this puzzle
        /// </summary>
        public int Column { get; private set; }

        /// <summary>
        /// Gets the absolute position of this puzzle.
        /// </summary>
        public int Position
        {
            get { return Row*Table.Width + Column; }
        }

        /// <summary>
        /// Creates a new puzzle for the specified table.
        /// </summary>
        /// <param name="table">Table managing the puzzle</param>
        /// <param name="puzzleId">Id of the puzzle</param>
        public PuzzlePiece(GameController table, int puzzleId)
        {
            this.InitializeComponent();
            Table = table;
            PuzzleId = puzzleId;
        }

        /// <summary>
        /// Gets or sets the identifier of the puzzle.
        /// </summary>
        public int PuzzleId
        {
            get { return _puzzleId; }
            set 
            {
                _puzzleId = value;
                CaptionText.Text = (value + 1).ToString();
            }
        }

        /// <summary>
        /// Sets the position of the puzzle as specified.
        /// </summary>
        /// <param name="row">Position row</param>
        /// <param name="column">Position column</param>
        public void SetPosition(int row, int column)
        {
            Column = column;
            SetValue(Canvas.LeftProperty, PUZZLE_WIDTH * column);
            Row = row;
            SetValue(Canvas.TopProperty, PUZZLE_HEIGTH * row);
        }

        /// <summary>
        /// Checks whether the puzzle can be moved ot the empty space in the table.
        /// </summary>
        public bool CanBeMoved
        {
            get { return Table.TargetPositionsFrom(Position).Any(pos => Table[pos] == null); }
        }

        /// <summary>
        /// Marks a puzzle pices when it's moveable
        /// </summary>
        private void border_PointerEntered(object sender, Windows.UI.Xaml.Input.PointerRoutedEventArgs e)
        {
            if (CanBeMoved)
            {
                VisualStateManager.GoToState(this, "PointerOver", true);
            }
        }

        /// <summary>
        /// Unmarks a puzzle piece
        /// </summary>
        private void border_PointerExited(object sender, Windows.UI.Xaml.Input.PointerRoutedEventArgs e)
        {
            VisualStateManager.GoToState(this, "Normal", true);
        }

        /// <summary>
        /// Moves a puzzle piece
        /// </summary>
        private void border_Tapped(object sender, TappedRoutedEventArgs e)
        {
            Table.MakeMove(Position);
            VisualStateManager.GoToState(this, "Normal", true);
        }
    }

}
